import ast
import copy
import glob
import json
import logging
import pathlib
import re
import tarfile
import tempfile
import zipfile
from functools import reduce
from shutil import copyfile
from typing import IO, Any, Callable, Dict, List, Optional, Tuple

from baseten_scaffolding.definitions.base import \
    AbstractModelScaffoldDefinition
from baseten_scaffolding.model_inference import (infer_pytorch_packages,
                                                 infer_sklearn_packages,
                                                 infer_tensorflow_packages)
from baseten_scaffolding.scaffold.build import scaffold, scaffold_custom
from colorama import Fore, Style

import baseten
from baseten.baseten_deployed_model import BasetenDeployedModel
from baseten.common import api
from baseten.common.core import (MODEL_FILENAME, FrameworkNotSupportedError,
                                 ModelClassImplementationError,
                                 ModelFilesMissingError, ModelInputShapeError,
                                 ModelNotSupportedError, Semver,
                                 raises_api_error)
from baseten.common.settings import (API_URL_BASE, CUSTOM, KERAS, PYTORCH,
                                     SKLEARN, TENSORFLOW)
from baseten.common.util import parse_requirements_file, zipdir

logger = logging.getLogger(__name__)


@raises_api_error
def deploy_model(serialized_model: IO[Any],
                 model_framework: str,
                 model_name: str = None,
                 model_type: str = None,
                 input_shape: List[int] = None,
                 feature_names: List[str] = None,
                 feature_summary: dict = None,
                 class_labels: List[str] = None,
                 version_bump: Semver = Semver.MINOR.value,
                 model_framework_req: List[str] = None,
                 model_files_dict: Dict = None) -> BasetenDeployedModel:
    """Create a model version based on the serialized_model. If baseten wasn't initialized with a model_id, then
    create a model first and then create the model version.

    Args:
        serialized_model (file): A file-like object that is the serialized representation of the model object.
        model_framework (str): The framework used to create the model.
        model_name (str, optional): The name of the model to be created, if necessary.
        model_type (str, optional): The type of the model to be created.
        input_shape (List[int], optional): The shape of the model's input; e.g: [10] for a model with 10 features;
            [28, 28] for a model with 2 input dimensions of size 28 each.
        feature_names (list, optional): The list of feature names.
        feature_summary (dict, optional): A dict of summary stats with feature names as keys.
        class_labels (List[str], optional): The list of target class labels, if applicable.
        version_bump (str, optional): The version bump for this deployment, one of 'MAJOR', 'MINOR', 'PATCH'.
        model_framework_req (List[str], optional): The model_framework specific requirements
    Returns:
        BasetenDeployedModel

    Raises:
        ApiError: If there was an error communicating with the server.
        ValueError: If no model was provided.
        FrameworkNotSupportedError: If provided model's framework is not supported.
    """
    if not serialized_model:
        raise ValueError('A model was not provided!')
    if not input_shape:
        input_shape = []
    if not feature_summary:
        feature_summary = {}
    if not class_labels:
        class_labels = []
    if not model_framework_req:
        model_framework_req = {}

    logger.info('👽 👋 Making contact with BaseTen 👋 👽')

    file_ext = _model_ext_from_framework(model_framework)
    model_key = api.upload_model(serialized_model, file_ext, MODEL_FILENAME)
    model_info = _exists_model(model_name, baseten.working_model_id, api.models)
    if not model_info:
        logger.debug('No model found with provided name/id.')
        logger.debug(f'Creating new model named {Fore.BLUE}{model_name}{Style.RESET_ALL}')
        model_and_version_json = api.create_model(model_name,
                                                  model_key,
                                                  model_framework,
                                                  model_type,
                                                  input_shape,
                                                  feature_names,
                                                  feature_summary,
                                                  class_labels,
                                                  model_framework_req,
                                                  version_bump,
                                                  model_files_dict)
        logger.debug(f'Created model:\n{json.dumps(model_and_version_json, indent=4)}')
        logger.info(f'Successfully created new model named {Fore.BLUE}{model_name}{Style.RESET_ALL}.')

        model_id = model_and_version_json['id']
        model_version_id = model_and_version_json['version_id']
        # Set the newly created model to be the working model for future commands
        baseten.working_model_id = model_id
    else:
        model_id, model_name = model_info
        logger.info(f"Model '{model_name}' found. Creating a new version.")
        model_version_json = api.create_model_version(model_id,
                                                      model_key,
                                                      model_framework,
                                                      model_type,
                                                      input_shape,
                                                      feature_names,
                                                      feature_summary,
                                                      class_labels,
                                                      model_framework_req,
                                                      version_bump,
                                                      model_files_dict)

        model_version_id = model_version_json['id']
        logger.info(
            f'Successfully created version {Fore.BLUE}{model_version_id}{Style.RESET_ALL} for {model_name}.')

    logger.debug(f'{Fore.BLUE}Deploying model version.')
    model_version_web_url = f'{API_URL_BASE}/models/{model_id}/versions/{model_version_id}'
    logger.info('🏁 Deploying model 🏁')
    visit_message = f'|  Visit {Fore.BLUE}{model_version_web_url}{Style.RESET_ALL} for deployment status  |'
    visit_message_len = len(visit_message) - len(Fore.BLUE) - len(Style.RESET_ALL)
    logger.info(''.join(['-' for _ in range(visit_message_len)]))
    logger.info(visit_message)
    logger.info(''.join(['-' for _ in range(visit_message_len)]))

    return BasetenDeployedModel(model_version_id=model_version_id, model_name=model_name)


def _exists_model(model_name: Optional[str],
                  model_id: Optional[str],
                  models_provider: Callable) -> Optional[Tuple[str, str]]:
    """Checks if an effective model exists for the purpose of deploy or one needs to be created.

    If model name is supplied and a model with that name exists then it's picked up. Otherwise
    working model id in the session is tried.

    Returns id of model if model exists, or None
    """
    models = models_provider()['models']
    model_id_by_name = {model['name']: model['id'] for model in models}
    model_name_by_id = {model['id']: model['name'] for model in models}

    if model_name:
        if model_name in model_id_by_name:
            return model_id_by_name[model_name], model_name
        else:
            return None

    # No model_name supplied, try to work with working model id
    if model_id:
        if model_id in model_name_by_id:
            return model_id, model_name_by_id[model_id]
        else:
            logger.warning('Working model id not found in deployed models')
            return None

    # No model name or working model
    return None


@raises_api_error
def deploy_custom_model(
    model_name: str,
    model_class: str,
    model_files: List[str],
    requirements_file: str = None,
    version_bump: Semver = Semver.MINOR.value,
    model_type: str = None,
) -> BasetenDeployedModel:
    """Deploy a custom model.

    Deploy a model class (`model_class`) which implements the functions `load` and `predict`. `load()` is
    called on the model to initialize it, so is the place to load from a serialized format, download necessary
    files, etc. The `predict` method is passed the raw json inputs to the model, and must return json
    serializable data. Any additional pre or post processing can be run before returning the response.

    Model files should be all contained within the current working directory, and will be packaged relative
    to the current working directory.

    See docs.baseten.co for further documentation and examples.

    Args:
        model_name (str): The name of the model to be created.
        model_class (str): The name of the model class which implements `load` and `predict` methods,
            defined in one of `model_files`.
        model_files (List[str]): A list of files to include with the model deployment, including serialized
            models, python files, data, embeddings, etc. Files will match based on pathname patterns, eg *.py
            will match all python files in the current working directory.

            Must include file defining the model class.
        requirements_file (str, optional): A requirements.txt file specifying requirements to be installed
            alongside the model.
        version_bump (Semver, optional): The version bump for this deployment, one of 'MAJOR', 'MINOR', 'PATCH'.
            Defaults to Semver.MINOR.value.

    Returns:
        BasetenDeployedModel
    """
    all_model_files = list(set([_file for f in model_files for _file in glob.glob(f)]))
    serialized_model, class_def_file = _serialize_custom_model(model_class, all_model_files)
    model_files_dict = {'class_def_file': class_def_file, 'model_class': model_class}
    if requirements_file:
        model_framework_req = parse_requirements_file(requirements_file)
    else:
        model_framework_req = {}
    if not model_type:
        model_type = CUSTOM

    deployed_model = deploy_model(serialized_model,
                                  CUSTOM,
                                  model_name,
                                  model_type,
                                  version_bump=version_bump,
                                  model_framework_req=model_framework_req,
                                  model_files_dict=model_files_dict)
    serialized_model.close()
    return deployed_model


def _legacy_model_framework(model_framework: str):
    # tf.Keras does not always resolve to tensorflow, in newer versions of TF (>2.3.4) it will show as keras
    # in the legacy world we still want this to be tensorflow
    if model_framework == KERAS:
        return TENSORFLOW
    return model_framework


@raises_api_error
def deploy_model_object(
    model: Any,
    model_name: str = None,
    model_files: List[str] = None,
    feature_names: List[str] = None,
    class_labels: List[str] = None,
    version_bump: Semver = Semver.MINOR.value,
) -> BasetenDeployedModel:
    """Create a model version based on the model. If baseten wasn't initialized with a model_id, then
    create a model first and then create the model version.

    Args:
        model (an in-memory model object): A model object to be deployed (e.g. a RandomForestClassifier object)
        model_name (str, optional): The name of the model to be created, if necessary.
        model_files (List[str], optional): Additional files required for model definition, e.g class
            definitions for pytorch models.
        feature_names (List[str], optional): The list of feature names. If not provided,
            a list of feature names that match the model's input dimensions will be automatically generated.
            If the model's input has multiple dimensions, feature_names should be a flattened list of names.
        class_labels (List[str], optional): The list of target class labels, if applicable.
        version_bump (str, optional): The version bump for this deployment, one of 'MAJOR', 'MINOR', 'PATCH'.
    Returns:
        BasetenDeployedModel

    Raises:
        FrameworkNotSupportedError: If the model framework used is not supported.
        ModelInputShapeError: feature_names does not match the model's signature.
        ModelFilesMissingError: model is missing required supporting files for deployment.
        ApiError: If there was an error communicating with the server.
    """
    if hasattr(model, 'baseten'):
        model = copy.deepcopy(model)
        del model.baseten
    logger.info('🤖 Extracting model metadata 🤖')
    model_class = model.__class__
    model_framework, _, _ = model_class.__module__.partition('.')
    model_framework = _legacy_model_framework(model_framework)

    if model_framework not in {SKLEARN, TENSORFLOW}:
        try:
            import torch
            if issubclass(model_class, torch.nn.Module):
                model_framework = PYTORCH
            else:
                raise FrameworkNotSupportedError(f'Models must be one of {SKLEARN}, {TENSORFLOW}, or {PYTORCH}.')
        except ModuleNotFoundError:
            raise FrameworkNotSupportedError(f'Models must be one of {SKLEARN}, {TENSORFLOW}, or {PYTORCH}.')

    model_type = model_class.__name__
    input_shape = _model_input_shape(model, model_framework)
    logger.info(f'Preparing {Fore.BLUE}{model_framework} - {model_type}{Style.RESET_ALL} model.')

    if feature_names:
        input_features_count = reduce(lambda x, y: x * y, input_shape)
        if len(feature_names) != input_features_count:
            raise ModelInputShapeError(f'The model has {input_features_count} features but feature_names provided '
                                       f'{len(feature_names)}')

    if version_bump not in [item.value for item in Semver]:
        raise ValueError(f'{version_bump} is not a valid semantic version bump. '
                         f'Must be one of MAJOR, MINOR, and PATCH')
    logger.info(f'Serializing {Fore.BLUE}{model_name}{Style.RESET_ALL}.')
    if model_framework == SKLEARN:
        serialized_model = _serialize_sklearn_model(model)
        model_framework_req = infer_sklearn_packages()
        deployed_model = deploy_model(serialized_model,
                                      model_framework,
                                      model_name,
                                      model_type,
                                      input_shape,
                                      feature_names=feature_names,
                                      class_labels=class_labels,
                                      version_bump=version_bump,
                                      model_framework_req=model_framework_req)
        serialized_model.close()
    elif model_framework == TENSORFLOW:
        serialized_model = _serialize_tf_model(model)
        model_framework_req = infer_tensorflow_packages()
        deployed_model = deploy_model(serialized_model,
                                      model_framework,
                                      model_name,
                                      model_type,
                                      input_shape,
                                      feature_names=feature_names,
                                      class_labels=class_labels,
                                      version_bump=version_bump,
                                      model_framework_req=model_framework_req)
        serialized_model.close()
    elif model_framework == PYTORCH:
        serialized_model, model_files_dict = _serialize_pytorch_model(model, model_files)
        model_framework_req = infer_pytorch_packages()
        deployed_model = deploy_model(serialized_model,
                                      model_framework,
                                      model_name,
                                      model_type,
                                      input_shape,
                                      feature_names=feature_names,
                                      class_labels=class_labels,
                                      version_bump=version_bump,
                                      model_framework_req=model_framework_req,
                                      model_files_dict=model_files_dict)
        serialized_model.close()

    return deployed_model


def build_baseten_scaffold(
    model: Any,
    model_name: str = None,
    model_files: List[str] = None,
    data_files: List[str] = None,
    target_directory: str = None,
    requirements_file: str = None,
):
    """Create a scaffold directory with the given model. A scaffold directory is a build context designed to be built
    as a container locally or uploaded into a baseten serving environment.

    Args:
        model (an in-memory model object): A model object to be deployed (e.g. a keras, sklearn, or pytorch model
            object)
        model_name (str, optional): The name of the model to be created, if necessary.
        model_files (List[str], optional): Additional files required for model definition, e.g class
            definitions for pytorch models or other python code. Can be a glob that resolves to files for the root
            directory or a directory path. Files and directories here will be on the Python path
        data_files (List[str], optional): Additional files required for model operation. Can be a glob that resolves to
            files for the root directory or a directory path.
        target_directory (str, optional): The local directory target for the scaffold. Otherwise a temporary directory
            will be generated
        requirements_file (str, optional): A file of packages in a PIP requirements format to be installed in the
            container environment.
    Returns:
        AbstractModelScaffoldDefinition
    """
    if not model_name:
        import coolname
        model_name = coolname.generate_slug(2)
    local_name = model_name.replace(' ', '_')

    b10_scaffold = scaffold(
        model=model,
        model_files=model_files,
        data_files=data_files,
        target_directory=target_directory,
        requirements_file=requirements_file,
    )
    logger.info(f'To build this model server locally execute '
                f'`{b10_scaffold.docker_build_string} -t {local_name.lower()}`')
    logger.info(f'To run this model server locally execute `docker run --rm -p 8080:8080 {local_name.lower()}`')
    logger.info(f'To use the Python shell locally execute `docker run --rm -it {local_name.lower()} python3`')
    return b10_scaffold


def build_baseten_custom_scaffold(
    model_name: str,
    model_files: List[str] = None,
    data_files: List[str] = None,
    target_directory: str = None,
    requirements_file: str = None,
    model_class: str = None,
):
    """
    Create a scaffold directory for a custom model. A scaffold directory is a build context designed to be built as a
    container locally or uploaded into a baseten serving environment.

        Args:
            model_name (str, optional): The name of the model to be created, if necessary.
            model_files (List[str], optional): Additional files required for model definition, e.g class
                definitions for pytorch models or other python code. Can be a glob that resolves to files for the root
                directory or a directory path. Files and directories here will be on the Python path
            data_files (List[str], optional): Additional files required for model operation. Can be a glob that resolves
                to files for the root directory or a directory path.
            target_directory (str, optional): The local directory target for the scaffold. Otherwise a temporary
                directory will be generated
            requirements_file (str, optional): A file of packages in a PIP requirements format to be installed in the
                container environment.
            model_class (str, optional): The python class name for a model object that exists as a code within
                `model_files`
        Returns:
            AbstractModelScaffoldDefinition
    """
    if not model_name:
        import coolname
        model_name = coolname.generate_slug(2)
    local_name = model_name.replace(' ', '_')

    b10_scaffold = scaffold_custom(
        model_files=model_files,
        data_files=data_files,
        target_directory=target_directory,
        requirements_file=requirements_file,
        model_class=model_class
    )
    logger.info(f'To build this model server locally execute '
                f'`{b10_scaffold.docker_build_string} -t {local_name.lower()}`')
    logger.info(f'To run this model server locally execute `docker run --rm -p 8080:8080 {local_name.lower()}`')
    logger.info(f'To use the Python shell locally execute `docker run --rm -it {local_name.lower()} python3`')
    logger.info('Be sure to read the `README.md` for guidance on construction of a working custom scaffold')
    return b10_scaffold


@raises_api_error
def deploy_baseten_scaffold(
    b10_scaffold: AbstractModelScaffoldDefinition,
    model_name: str = None,
    version_bump: Semver = Semver.MINOR.value,
):
    """
    Given an existing baseten scaffold object, deploy it onto the baseten infastructure.

    Args:
        b10_scaffold: A scaffold object representing a local baseten_scaffold directory
        model_name (str, optional): The name of the model to be created, if necessary.
        version_bump (str, optional): The version bump for this deployment, one of 'MAJOR', 'MINOR', 'PATCH'.

    Returns:
        BasetenDeployedModel
    """
    logger.info(f'Serializing {Fore.BLUE}{model_name}{Style.RESET_ALL} scaffold.')
    temp_file = tempfile.NamedTemporaryFile(suffix='.tgz')
    with tarfile.open(temp_file.name, 'w:gz') as tar:
        tar.add(b10_scaffold.scaffold_dir, arcname='.')
    temp_file.file.seek(0)

    model_info = _exists_model(model_name, baseten.working_model_id, api.models)

    logger.info('Making contact with BaseTen 👋 👽')
    s3_key = api.upload_model(temp_file, 'tgz', MODEL_FILENAME)
    if not model_info:
        model_version_json = api.create_model_from_scaffold(
            model_name,
            s3_key,
            b10_scaffold.model_framework,
            b10_scaffold.model_type,
            b10_scaffold.model_framework_requirements,
            b10_scaffold.dockerfile,
            b10_scaffold.build_args,
            version_bump
        )
        logger.info(f'Created model:\n{json.dumps(model_version_json, indent=4)}')
        logger.info(f'Successfully registered model {Fore.BLUE}{model_name}{Style.RESET_ALL}.')

        model_id = model_version_json['id']
        model_version_id = model_version_json['version_id']
        # Set the newly created model model_version_json be the working model for future commands
        baseten.working_model_id = model_id
    else:
        model_id, model_name = model_info
        logger.info(f"Model '{model_name}' found. Creating a new version.")

        model_version_json = api.create_model_version_from_scaffold(model_id,
                                                                    s3_key,
                                                                    b10_scaffold.model_framework,
                                                                    b10_scaffold.model_type,
                                                                    b10_scaffold.model_framework_requirements,
                                                                    b10_scaffold.dockerfile,
                                                                    b10_scaffold.build_args,
                                                                    version_bump)

        model_version_id = model_version_json['id']
        logger.info(
            f'Successfully created version {Fore.BLUE}{model_version_id}{Style.RESET_ALL} for {model_name}.')

    logger.info(f'{Fore.BLUE}Deploying model version.')
    model_version_web_url = f'{API_URL_BASE}/models/{model_id}/versions/{model_version_id}'
    logger.info('🏁 The model is being built and deployed right now 🏁')
    visit_message = f'|  Visit {Fore.BLUE}{model_version_web_url}{Style.RESET_ALL} for deployment status  |'
    visit_message_len = len(visit_message) - len(Fore.BLUE) - len(Style.RESET_ALL)
    logger.info(''.join(['-' for _ in range(visit_message_len)]))
    logger.info(visit_message)
    logger.info(''.join(['-' for _ in range(visit_message_len)]))
    return BasetenDeployedModel(model_version_id=model_version_id, model_name=model_name)


def build_and_deploy_baseten_scaffold(
    model: Any,
    model_name: str = None,
    model_files: List[str] = None,
    data_files: List[str] = None,
    target_directory: str = None,
    requirements_file: str = None,
    version_bump: Semver = Semver.MINOR.value,
):
    """Create a scaffold directory with the given model and upload it to the basten serving environment. A scaffold
    directory is a build context designed to be built as a container locally or uploaded into a baseten serving
    environment.

    Args:
        model (an in-memory model object): A model object to be deployed (e.g. a keras, sklearn, or pytorch model
            object)
        model_name (str, optional): The name of the model to be created, if necessary.
        model_files (List[str], optional): Additional files required for model definition, e.g class
            definitions for pytorch models or other python code. Can be a glob that resolves to files for the root
            directory or a directory path. Files and directories here will be on the Python path
        data_files (List[str], optional): Additional files required for model operation. Can be a glob that resolves to
            files for the root directory or a directory path.
        target_directory (str, optional): The local directory target for the scaffold. Otherwise a temporary directory
            will be generated
        requirements_file (str, optional): A file of packages in a PIP requirements format to be installed in the
            container environment.
        version_bump (str, optional): The version bump for this deployment, one of 'MAJOR', 'MINOR', 'PATCH'.
    Returns:
        BasetenDeployedModel
    """

    b10_scaffold = build_baseten_scaffold(
        model, model_name, model_files, data_files, target_directory, requirements_file
    )
    return deploy_baseten_scaffold(
        b10_scaffold,
        model_name,
        version_bump
    )


def build_and_deploy_baseten_custom_scaffold(
    model_name: str = None,
    model_files: List[str] = None,
    data_files: List[str] = None,
    target_directory: str = None,
    requirements_file: str = None,
    model_class: str = None,
    version_bump: Semver = Semver.MINOR.value,
):
    """
    Create a scaffold directory for a custom model and upload it into the baseten serving environment.

        Args:
            model_name (str, optional): The name of the model to be created, if necessary.
            model_files (List[str], optional): Additional files required for model definition, e.g class
                definitions for pytorch models or other python code. Can be a glob that resolves to files for the root
                directory or a directory path. Files and directories here will be on the Python path
            data_files (List[str], optional): Additional files required for model operation. Can be a glob that resolves
                to files for the root directory or a directory path.
            target_directory (str, optional): The local directory target for the scaffold. Otherwise a temporary
                directory will be generated
            requirements_file (str, optional): A file of packages in a PIP requirements format to be installed in the
                container environment.
            model_class (str, optional): The python class name for a model object that exists as a code within
                `model_files`
            version_bump (str, optional): The version bump for this deployment, one of 'MAJOR', 'MINOR', 'PATCH'.
    Returns:
        BasetenDeployedModel
    """

    b10_scaffold = build_baseten_custom_scaffold(
        model_name, model_files, data_files, target_directory, requirements_file, model_class
    )
    return deploy_baseten_scaffold(
        b10_scaffold,
        model_name,
        version_bump
    )


def _serialize_sklearn_model(model: Any) -> tempfile.SpooledTemporaryFile:
    """Serializes a model based on the scikit-learn framework.

    Args:
        model (Any): The model object.

    Returns:
        tempfile.SpooledTemporaryFile: A temporary file wrapper.
    """
    import joblib
    model_joblib = tempfile.TemporaryFile()
    joblib.dump(model, model_joblib, compress=True)
    model_joblib.seek(0)
    return model_joblib


def _serialize_tf_model(model: Any) -> tempfile.NamedTemporaryFile:
    """Serializes a Tensorflow model as a zipped SavedModel.

    Args:
        model (Any): The model object.

    Returns:
        tempfile.NamedTemporaryFile: A temporary zipfile.ZipFile wrapper.
    """
    temp_model_dir = tempfile.TemporaryDirectory()
    temp_file = tempfile.NamedTemporaryFile(suffix='.zip')
    model.save(temp_model_dir.name)
    zip_file = zipfile.ZipFile(temp_file.name, 'w', zipfile.ZIP_DEFLATED)
    zipdir(temp_model_dir.name, zip_file)
    zip_file.close()
    temp_model_dir.cleanup()
    temp_file.file.seek(0)
    return temp_file


def _serialize_pytorch_model(model: Any, model_files: List[str] = None) -> Tuple[tempfile.NamedTemporaryFile, Dict]:
    """Serializes a PyTorch model as a zipped .pt file.

    Args:
        model (Any): The model object.

    Returns:
        tempfile.NamedTemporaryFile: A temporary zipfile.ZipFile wrapper.
    """
    model_class = model.__class__
    class_name = model_class.__name__
    if not model_files:
        raise ModelFilesMissingError(f'The file defining the PyTorch model class `{class_name}` is missing.')
    temp_model_dir = tempfile.TemporaryDirectory()
    temp_file = tempfile.NamedTemporaryFile(suffix='.zip')
    import torch
    torch.save(model.state_dict(), f'{temp_model_dir.name}/model.pt')
    has_supplied_model_class_definion = False
    model_files_dict = {'files': {}, 'model_class': class_name}
    for filepath in model_files:
        with open(filepath, 'r') as _file:
            file_contents = _file.read()
            model_files_dict['files'][filepath] = file_contents
            parsed_contents = ast.parse(file_contents)
            is_model_class_definion_file = (
                True if [
                    stmt for stmt in parsed_contents.body
                    if type(stmt) == ast.ClassDef
                    and stmt.name == class_name
                ] else False)
            if is_model_class_definion_file:
                model_files_dict['class_def_file'] = filepath
            has_supplied_model_class_definion = has_supplied_model_class_definion or is_model_class_definion_file
    if not has_supplied_model_class_definion:
        raise ModelFilesMissingError(f'The file defining the PyTorch model class `{class_name}` is missing.')

    zip_file = zipfile.ZipFile(temp_file.name, 'w', zipfile.ZIP_DEFLATED)
    zipdir(temp_model_dir.name, zip_file)
    zip_file.close()
    temp_model_dir.cleanup()
    temp_file.file.seek(0)
    return temp_file, model_files_dict


def _serialize_custom_model(class_name: str, model_files: List[str]) -> Tuple[tempfile.NamedTemporaryFile, str]:
    """Serializes a list of model files as a zip file.

    Args:
        class_name (str): The class_name for the model.
        model_files List[str]): A list of files to be packaged as the model deployment.

    Returns:
        tempfile.NamedTemporaryFile: A temporary zipfile.ZipFile wrapper.

    Raises:
        ModelFilesMissingError: If a file defining the model class is not supplied.
        ModelClassImplementationError: If the model class does not implement the required `load` and
            `predict` methods.

    Returns:
        (tempfile.NamedTemporaryFile, str): A tuple with a temporary zipfile.ZipFile wrapper and the filepath
            of the file defining the model class.
    """
    if not model_files:
        raise ModelFilesMissingError(f'The file defining the model class `{class_name}` is missing.')
    has_supplied_model_class_definion = False
    python_files = [f for f in model_files if f.endswith('.py')]
    class_def_file = None
    for filepath in python_files:
        with open(filepath, 'r') as _file:
            file_contents = _file.read()
            parsed_contents = ast.parse(file_contents)
            model_class_definion_file = [
                stmt for stmt in parsed_contents.body
                if type(stmt) == ast.ClassDef
                and stmt.name == class_name
            ]
            if model_class_definion_file:
                cls_function_names = [stmt.name for stmt in model_class_definion_file[0].body]
                if 'load' not in cls_function_names or 'predict' not in cls_function_names:
                    raise ModelClassImplementationError(f'The model class in {filepath} does not \
                        implement the required `load` and `predict` methods.')
                class_def_file = filepath
                has_supplied_model_class_definion = True
                break
    if not has_supplied_model_class_definion:
        raise ModelFilesMissingError(f'The file defining the model class `{class_name}` is missing.')
    temp_model_dir = tempfile.TemporaryDirectory()
    temp_file = tempfile.NamedTemporaryFile(suffix='.zip')
    for _f in model_files:
        file_name = pathlib.PurePath(_f).name
        path = pathlib.PurePath(temp_model_dir.name, file_name)
        copyfile(_f, path)
    zip_file = zipfile.ZipFile(temp_file.name, 'w', zipfile.ZIP_DEFLATED)
    zipdir(temp_model_dir.name, zip_file)
    zip_file.close()
    temp_model_dir.cleanup()
    temp_file.file.seek(0)
    return temp_file, class_def_file


def _model_ext_from_framework(model_framework: str) -> str:
    if model_framework == SKLEARN:
        return 'joblib'
    elif model_framework == TENSORFLOW:
        return 'zip'
    elif model_framework == PYTORCH:
        return 'zip'
    elif model_framework == CUSTOM:
        return 'zip'
    else:
        raise FrameworkNotSupportedError(f'Models derived from {model_framework} are not currently supported.')


def _model_input_shape(model, framework: str) -> List[int]:
    """Returns the shape of the model's input as an array; e.g: [10] for a model with 10 features; [28, 28] for a
    model with 2 input dimensions of size 28 each.
    """
    if framework == SKLEARN:
        n_features = _hack_n_features_from_sklearn_model(model)
        if n_features is None:
            return []
        return [n_features]
    elif framework == TENSORFLOW:
        input_shape = model.input_shape
        if isinstance(input_shape, list):
            raise ModelNotSupportedError("Baseten currently supports only single input models.")
        return list(input_shape[1:])  # The first dimension is None (batch size)
    elif framework == PYTORCH:
        return []
    else:
        raise FrameworkNotSupportedError(f'Models derived from {framework} are not currently supported.')


def _hack_n_features_from_sklearn_model(model) -> int:
    """Determines the size of the input for the given model.

    Due to inconsistent representation of feature inputs in sklearn, various methods/props
    are required to extract the input dimension based on the model type. It's a hack.

    Args:
        model: A sklearn model object.

    Returns:
        int: The number of features used to train the model.
    """
    if 'sklearn.neighbors' in str(model.__class__):
        return model._tree.data.shape[-1]
    if 'sklearn.naive_bayes.GaussianNB' in str(model.__class__):
        return model.theta_.shape[-1]
    if model.__module__ == 'sklearn.naive_bayes':
        return model.feature_log_prob_.shape[-1]
    if model.__module__ == 'sklearn.discriminant_analysis':
        return model.means_.shape[-1]
    if 'sklearn.gaussian_process' in model.__module__:
        return model.base_estimator_.X_train_.shape[-1]
    try:
        model.predict([[1]])
        n_features = 1
    except ValueError as e:
        # For many models (not the ones above), the correct input feature length is raised in the exception
        numbers = re.findall(r' (\d+)', str(e))
        n_features = [int(x) for x in numbers if x != '1'][-1]
    except Exception as e:
        logger.debug(f'Unable to determine input_shape because of: {e}, but not stopping 🤞')
        # we can't determine the input shape, but we don't want to stop the user from uploading
        n_features = None
    return n_features



"""Baseten

    isort:skip_file
"""

__version__ = '0.1.20'

import logging
from collections import defaultdict

from baseten import baseten_artifact

patched_modules = defaultdict(lambda: [])

from baseten.baseten_deployed_model import BasetenDeployedModel, models_summary  # noqa: E402
from baseten.baseten_deployed_model import install_requirements, requirements_status  # noqa: E402, F401
from baseten.common.model_deployer import deploy_custom_model, deploy_model_object, \
    build_and_deploy_baseten_scaffold, build_and_deploy_baseten_custom_scaffold, build_baseten_scaffold, \
    build_baseten_custom_scaffold, deploy_baseten_scaffold  # noqa: E402
from baseten.common.util import setup_logger  # noqa: E402
from baseten.common import settings  # noqa: E402


logger = logging.getLogger(__name__)
setup_logger('baseten', logging.INFO)
if settings.DEBUG:
    setup_logger('baseten', logging.DEBUG)
logger.debug(f'Starting the client with the server URL set to {settings.API_URL_BASE}')

# The Baseten model ID that either a) the user initialized baseten with, or b) was created when deploying a model.
working_model_id = None

deploy = build_and_deploy_baseten_scaffold  # This allows the user to call baseten.deploy(model)
deploy_custom = build_and_deploy_baseten_custom_scaffold  # This allows the user to call baseten.deploy_custom(...)

build_scaffold = build_baseten_scaffold
build_custom_scaffold = build_baseten_custom_scaffold
deploy_scaffold = deploy_baseten_scaffold

deploy_legacy = deploy_model_object  # This allows the user to call baseten.deploy_legacy(model)
deploy_custom_legacy = deploy_custom_model  # This allows the user to call baseten.deploy_custom_legacy(...)

create_artifact = baseten_artifact.create_artifact

models_summary = models_summary


def set_log_level(level):
    setup_logger('baseten', level)


def init(baseten_model_id=None):
    """Initialize Baseten

    Args:
        baseten_model_id (str, optional): The BaseTen model id to initialize the client with.
        If not provided, a new Baseten model will be created at deploy() time.
    """
    global working_model_id
    working_model_id = baseten_model_id


def deployed_model_id(model_id: str) -> BasetenDeployedModel:
    """Returns a BasetenDeployedModel object for interacting with the model model_id.

    Args:
        model_id (str)

    Returns:
        BasetenDeployedModel
    """
    return BasetenDeployedModel(model_id=model_id)


def deployed_model_version_id(model_version_id: str) -> BasetenDeployedModel:
    """Returns a BasetenDeployedModel object for interacting with the model version model_version_id.

    Args:
        model_version_id (str)

    Returns:
        BasetenDeployedModel
    """
    return BasetenDeployedModel(model_version_id=model_version_id)


def deployed_artifact_id(artifact_id: str):
    """Returns a BasetenArtifact object for interacting with given the artifact_id

    Args:
        artifact_id (str)

    Returns:
        BasetenArtifact
    """
    return baseten_artifact.BasetenArtifact(artifact_id=artifact_id)


def login(api_key: str):
    """Set the API key for the client.

    Args:
        api_key (str)
    """
    settings.set_config_value('api', 'api_key', api_key)
    logger.info('API key set.')



import json
import pathlib
from typing import List, Tuple

try:
    import numpy as np
    NUMPY_LIB = True
except ModuleNotFoundError:
    NUMPY_LIB = False
try:
    import pandas as pd
    PANDAS_LIB = True
except ModuleNotFoundError:
    PANDAS_LIB = False
try:
    import h5py
    H5PY_LIB = True
except ModuleNotFoundError:
    H5PY_LIB = False


NUMPY_NOT_FOUND_ERRROR_MSG = 'The library was not able to successfully import "numpy" package, check your Python ' \
                             'environment'
H5PY_LIB = 'The library was not able to successfully import "h5py" package, check your Python environment'


def coerce_data_as_numpy_array(data):
    """Validates that data can be coerced into a numpy array

    Args:
        data (Union[np.ndarray, pd.DataFrame, List]): The data to be transformed.

    Raises:
        TypeError: If data is wrong type.

    Returns:
        np.ndarray: A numpy array of data.
    """
    allowed_input = False
    if not NUMPY_LIB:
        raise ModuleNotFoundError(NUMPY_NOT_FOUND_ERRROR_MSG)
    elif NUMPY_LIB and isinstance(data, np.ndarray):
        allowed_input = True
    elif PANDAS_LIB and isinstance(data, pd.DataFrame):
        allowed_input = True
    elif isinstance(data, list):
        allowed_input = True
    if not allowed_input:
        raise TypeError(f'Data must be one of type [np.ndarray, pd.DataFrame, list], got {type(data)}')
    return np.array(data)


def coerce_input_to_json(inputs, metadata=None) -> Tuple[List, List]:
    """Takes supported inputs and coerces them into JSON

    Args:
        inputs: The data representing one or more inputs to call the model with.
                Accepted types are: list, pandas.DataFrame, and numpy.ndarray
            metadata (Union[pd.DataFrame, List[Dict]]): Metadata key/value pairs (e.g. name, url), one for each input.

    Raises:
        TypeError: If data is wrong type.

    Returns:
        Tuple[List, List] - the representation as JSON serializable lists
    """

    if PANDAS_LIB and isinstance(inputs, pd.DataFrame):
        inputs_list = inputs.to_dict('records')
    elif NUMPY_LIB and isinstance(inputs, np.ndarray):
        inputs_np_array = np.array(inputs)
        inputs_list = inputs_np_array.tolist()
    elif not isinstance(inputs, list):
        raise TypeError('predict can be called with either a list, a pandas DataFrame, or a numpy array.')
    else:
        inputs_list = inputs

    if PANDAS_LIB and isinstance(metadata, pd.DataFrame):
        metadata = metadata.to_dict(orient='records')
    return inputs_list, metadata


def build_h5_data_object(feature_data, target_data, metadata, data_temp_directory) -> str:
    """Creates an h5 file from the data object

    Args:
        feature_data (Union[np.ndarray, pd.DataFrame, List[List]]): The feature data to upload.
        target_data (Union[np.ndarray, pd.DataFrame, List[List]]): The target data to upload.
        metadata (List[Dict]): Metadata key/value pairs for the dataset.
        data_temp_directory (str): Folder name of a temporary directory to write `tmp.h5` to

        Returns:
            A filepath to the local h5 object
     """
    if not H5PY_LIB:
        raise ModuleNotFoundError(H5PY_LIB)
    if not NUMPY_LIB:
        raise ModuleNotFoundError(NUMPY_NOT_FOUND_ERRROR_MSG)
    data_temp_file = pathlib.PurePath(data_temp_directory, 'tmp.h5')
    feature_data_np = coerce_data_as_numpy_array(feature_data)
    data = {'features': feature_data_np}
    if np.any(target_data):
        data['targets'] = coerce_data_as_numpy_array(target_data)
    if np.any(metadata):
        data['metadata'] = json.dumps(metadata)

    h5_data = h5py.File(data_temp_file, mode='w')
    h5_sample_data_group = h5_data.create_group('sample_data')
    for key, np_obj in data.items():
        if key == 'metadata':
            dt = h5py.string_dtype(encoding='utf-8')
            h5_sample_data_group.create_dataset(key, data=np_obj, dtype=dt)
        else:
            h5_sample_data_group.create_dataset(key, data=np_obj)
    h5_data.close()
    return data_temp_file



"""Console script for baseten."""
import configparser
import functools
import logging
import sys
from urllib.parse import urlparse

import click

from baseten.common.settings import read_config, set_config_value
from baseten.common.util import setup_logger

logger = logging.getLogger(__name__)


def ensure_login(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        config = read_config()
        try:
            config.get('api', 'api_key')
        except configparser.NoOptionError:
            click.echo('You must first run the `baseten login` cli command.')
            sys.exit()
        result = func(*args, **kwargs)
        return result
    return wrapper


@click.group()
def cli_group():
    setup_logger('baseten', logging.INFO)


@cli_group.command()
@click.option('--server_url', prompt='BaseTen server URL')
def configure(server_url):
    if not _is_valid_url(server_url):
        click.echo('That is not a valid URL.')
        return
    set_config_value('api', 'url_base', server_url)
    click.echo('Saved server URL.')


@cli_group.command()
@click.option('--api_key', prompt='BaseTen API key', hide_input=True)
def login(api_key):
    set_config_value('api', 'api_key', api_key)
    click.echo('Saved API key.')


def _is_valid_url(url: str) -> bool:
    try:
        result = urlparse(url)
        return all([result.scheme, result.netloc])
    except ValueError:
        return False


if __name__ == '__main__':
    cli_group()



import functools
from enum import Enum

from requests import RequestException


class Error(Exception):
    """Base Baseten Error"""
    def __init__(self, message):
        super().__init__(message)
        self.message = message


class AuthorizationError(Error):
    """Raised in places where the user needs to be logged in and is not."""
    pass


class FrameworkNotSupportedError(Error):
    """Raised in places where the user attempts to use Baseten with an unsupported framework"""
    pass


class ModelNotSupportedError(Error):
    """Raised when a user attempts to use unsupported models for a
    framework that is otherwise supported"""
    pass


class ModelFilesMissingError(Error):
    """Raised when a user attempts to deploy a model which requires supporting files."""
    pass


class ModelClassImplementationError(Error):
    """Raised when a user attempts to deploy a model class which does not implement necessary methods."""
    pass


class ModelInputShapeError(Error):
    """Raised when deploying a model whose input shape doesn't match the number of feature names given by the user."""
    pass


class SampleDataInputShapeError(Error):
    """Raised when deploying a model whose input shape doesn't match the number of feature names given by the user."""
    pass


class ApiError(Error):
    """Errors in calling the Baseten API."""
    def __init__(self, message, response=None):
        super().__init__(message)
        self.response = response

    def __str__(self):
        error_str = self.message
        if self.response is not None:  # non-200 Response objects are falsy, hence the not None.
            error_str = f'{error_str}\n<Server response: {self.response.content}>'
        return error_str


def raises_api_error(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except RequestException as err:
            raise ApiError(f'Error calling {func.__name__}.', response=err.response) from err
    return wrapper


class Semver(Enum):
    MAJOR = 'MAJOR'
    MINOR = 'MINOR'
    PATCH = 'PATCH'


MODEL_FILENAME = 'model'
ARTIFACT_FILENAME = 'artifact'



import configparser
import os
from pathlib import Path

SKLEARN = 'sklearn'
KERAS = 'keras'
TENSORFLOW = 'tensorflow'
PYTORCH = 'pytorch'
CUSTOM = 'custom'

DEBUG = False
CONFIG_FILE_PATH = f'{Path.home()}/.baseten_config'


def is_config_file_valid() -> bool:
    try:
        config = configparser.ConfigParser()
        config.read(CONFIG_FILE_PATH)
        # read() doesn't raise an exception if the file doesn't exist
        return os.path.exists(CONFIG_FILE_PATH)
    except configparser.ParsingError:
        return False


def set_config_defaults():
    config = configparser.ConfigParser()
    if is_config_file_valid():
        config.read(CONFIG_FILE_PATH)
    if not config.has_section('api'):
        config.add_section('api')
    config.set('api', 'url_base', 'https://app.baseten.co')
    with open(CONFIG_FILE_PATH, 'w') as configfile:
        config.write(configfile)


def read_config():
    if not is_config_file_valid():
        set_config_defaults()
    config = configparser.ConfigParser()
    config.read(CONFIG_FILE_PATH)
    return config


def set_config_value(section: str, key: str, value: str):
    config = read_config()
    if section not in config.sections():
        config.add_section(section)
    config.set(section, key, value)
    with open(CONFIG_FILE_PATH, 'w') as configfile:
        config.write(configfile)


def _api_url_base() -> str:
    config = read_config()
    try:
        return config.get('api', 'url_base')
    except configparser.NoSectionError:
        set_config_defaults()
        return config.get('api', 'url_base')


API_URL_BASE = _api_url_base()



import base64
import copy
import json
import logging
import os
import pathlib

import pkg_resources
from colorama import Fore, Style
from pkg_resources.extern.packaging.requirements import InvalidRequirement

# list from https://scikit-learn.org/stable/developers/advanced_installation.html
SKLEARN_REQ_MODULE_NAME = {
    'numpy',
    'scipy',
    'joblib',
    'scikit-learn',
    'threadpoolctl',
}

# list from https://www.tensorflow.org/install/pip
# if problematic, lets look to https://www.tensorflow.org/install/source
TENSORFLOW_REQ_MODULE_NAME = {
    'tensorflow',
}


# list from https://pytorch.org/get-started/locally/
PYTORCH_REQ_MODULE_NAME = {
    'torch',
    'torchvision',
    'torchaudio',
}


LOG_COLORS = {
    logging.ERROR: Fore.RED,
    logging.DEBUG: Fore.MAGENTA,
    logging.WARNING: Fore.YELLOW,
    logging.INFO: Fore.GREEN,
}


class ColorFormatter(logging.Formatter):
    def format(self, record, *args, **kwargs):
        new_record = copy.copy(record)
        if new_record.levelno in LOG_COLORS:
            new_record.levelname = "{color_begin}{level}{color_end}".format(
                level=new_record.levelname,
                color_begin=LOG_COLORS[new_record.levelno],
                color_end=Style.RESET_ALL,
            )
        return super(ColorFormatter, self).format(new_record, *args, **kwargs)


def setup_logger(package_name, level):
    baseten_logger = logging.getLogger(package_name)
    baseten_logger.setLevel(level)
    handler = logging.StreamHandler()
    formatter = ColorFormatter(fmt='%(levelname)s %(message)s')
    handler.setFormatter(formatter)
    if baseten_logger.hasHandlers():
        baseten_logger.handlers.clear()
    baseten_logger.addHandler(handler)


def base64_encoded_json_str(obj):
    return base64.b64encode(str.encode(json.dumps(obj))).decode('utf-8')


def zipdir(path, zip_handler):
    for root, dirs, files in os.walk(path):
        relative_root = ''.join(root.split(path))
        for _file in files:
            zip_handler.write(
                os.path.join(root, _file), os.path.join(f'model{relative_root}', _file))


def print_error_response(response):
    print(Fore.YELLOW + f'{response["message"]}')
    print('---------------------------------------------------------------------------')
    print(Fore.GREEN + 'Stack Trace:')
    if 'exception' in response:
        excp = response['exception']
        for excp_st_line in excp['stack_trace']:
            print(Fore.GREEN + '---->' + Fore.WHITE + f'{excp_st_line}')
        print(Fore.RED + f'Exception: {excp["message"]}')


def parse_requirements_file(requirements_file: str) -> dict:
    name_to_req_str = {}
    with pathlib.Path(requirements_file).open() as reqs_file:
        for raw_req in reqs_file.readlines():
            try:
                req = pkg_resources.Requirement.parse(raw_req)
                if req.specifier:
                    name_to_req_str[req.name] = str(req)
                else:
                    name_to_req_str[str(req)] = str(req)
            except InvalidRequirement:
                # there might be pip requirements that do not conform
                raw_req = str(raw_req).strip()
                name_to_req_str[f'custom_{raw_req}'] = raw_req
            except ValueError:
                # can't parse empty lines
                pass

    return name_to_req_str



import json
import logging
import tempfile
import time
from typing import Dict, List

import requests
import yaml

from baseten.common import api
from baseten.common.core import raises_api_error
from baseten.common.lib_support import (build_h5_data_object,
                                        coerce_input_to_json)

logger = logging.getLogger(__name__)

REQUIREMENTS_INSTALLATION_STATUS_RETRY_INTERVAL_SEC = 3
REQUIREMENTS_INSTALLATION_STATUS_MAX_TRIES = 20


@raises_api_error
def install_requirements(req_filepath: str):
    with open(req_filepath, 'r') as fp:
        requirements_txt = fp.read()
    logger.info('🚀 Sending requirements to BaseTen 🚀')
    resp = api.install_requirements(requirements_txt)
    status = resp['status']
    if status == 'PROCESSING':
        logger.info('🐳 Requirements are being installed 🐳')

        requirement_id = resp['id']
        tries = 0
        while tries < REQUIREMENTS_INSTALLATION_STATUS_MAX_TRIES:
            time.sleep(REQUIREMENTS_INSTALLATION_STATUS_RETRY_INTERVAL_SEC)
            resp = api.requirement_status(requirement_id)
            status = resp['status']
            if status != 'PROCESSING':
                break
            tries += 1
        else:
            logger.info('⌛ Requirements are still being installed. Check the status by running '
                        f'baseten.requirements_status(\'{requirement_id}\') ⌛')
    if status == 'SUCCEEDED':
        logger.info('🖖 Installed requirements successfully 🖖')
    elif status == 'FAILED':
        error_message = resp['error_message']
        logger.info(f'⚠️ Failed to install requirements. Error: "{error_message}" ⚠️')


@raises_api_error
def requirements_status(requirement_id: str):
    return api.requirement_status(requirement_id)


class BasetenDeployedModel:
    """A model backed by baseten serving. Provide either the model_id or the model_version_id."""
    def __init__(self, model_id: str = None,
                 model_version_id: str = None,
                 model_name: str = None):
        if not model_id and not model_version_id:
            raise ValueError('Either model_id or model_version_id must be provided.')

        if model_id and model_version_id:
            raise ValueError('Must provide either model_id or model_version_id; not both.')

        self._model_id = model_id
        self._model_version_id = model_version_id
        self._model_name = model_name
        self._model_config = None

    @property
    def model_version_id(self):
        return self._model_version_id

    @raises_api_error
    def predict(self,
                inputs,
                metadata=None) -> List[List]:
        """Invokes the model given the input dataframe.

        Args:
            inputs: The data representing one or more inputs to call the model with.
                Accepted types are: list, pandas.DataFrame, and numpy.ndarray
            metadata (Union[pd.DataFrame, List[Dict]]): Metadata key/value pairs (e.g. name, url), one for each input.

        Returns:
            A list of inferences for each given input; e.g.: [[3], [9]] would indicate the prediction for the
            first input in inputs_df is [3], and the prediction for the second is [9].

        Raises:
            TypeError: If the provided inputs is not of a supported type.
            ApiError: If there was an error communicating with the server.
        """

        inputs_list, metadata = coerce_input_to_json(inputs, metadata)

        if self._model_version_id:
            return api.predict_for_model_version(self._model_version_id, inputs_list, metadata)
        return api.predict_for_model(self._model_id, inputs_list, metadata)

    @raises_api_error
    def update_model_features(self, model_config_file_path: str):
        """Update the model's feature names and output class labels (if any) based on the config
        found at `model_config_file_path`

        Args:
            model_config_file_path (str): The path to the model config file
        """
        config_yaml = yaml.safe_load(open(model_config_file_path, 'r'))
        feature_names = list(config_yaml['model_features']['features'])
        class_labels = config_yaml.get('model_class_labels', [])
        api.update_model_features(self._model_version_id, feature_names, class_labels)

    @raises_api_error
    def set_primary(self):
        """Promote this version of the model as the primary version.
        Raises:
            ApiError: If there was an error communicating with the server.
        """
        if not self._model_version_id:
            raise ValueError('Only a BasetenDeployedModel backed by a model_version can be set as primary.')
        return api.set_primary(self._model_version_id)

    @raises_api_error
    def upload_sample_data(self,
                           feature_data,
                           target_data=None,
                           metadata: List[Dict] = None,
                           data_name: str = 'validation_data',
                           ) -> Dict:
        """Upload a subset of the training/validation data to be used for
            - Summary statistics for the model
            - To detect model drift
            - To use as baseline data for model interpretability
            - To seed new data in the client.

        Training and validation data with targets must be uploaded with the targets separate.

        Args:
            feature_data (Union[np.ndarray, pd.DataFrame, List[List]]): The feature data to upload.
            target_data (Union[np.ndarray, pd.DataFrame, List[List]]): The target data to upload.
            metadata (List[Dict]): Metadata key/value pairs for the dataset.
            data_name (str): The name of the data set.

        Returns:
            Dict: The status of the upload.
        """
        if not self._model_version_id:
            raise ValueError('Please use on a BasetenDeployedModel instantiated with a model_version_id.')

        signed_s3_upload_post = api.signed_s3_upload_post(data_name)
        logger.debug(f'Signed s3 upload post:\n{json.dumps(signed_s3_upload_post, indent=4)}')

        with tempfile.TemporaryDirectory() as data_temp_directory:
            data_temp_file = build_h5_data_object(feature_data, target_data, metadata, data_temp_directory)
            files = {'file': (f'{data_name}.h5', open(data_temp_file, 'rb'))}
            form_fields = signed_s3_upload_post['form_fields']
            form_fields['AWSAccessKeyId'] = form_fields.pop('aws_access_key_id')
            s3_key = form_fields['key']
            requests.post(signed_s3_upload_post['url'], data=form_fields, files=files)
        return api.register_data_for_model(s3_key, self.model_version_id, data_name)

    @raises_api_error
    def artifacts(self):
        from baseten.baseten_artifact import BasetenArtifact
        if self._model_version_id:
            return [
                BasetenArtifact(**instance) for instance in api.model_version_linked_artifacts(self._model_version_id)
            ]
        return [BasetenArtifact(**instance) for instance in api.model_linked_artifacts(self._model_id)]

    def __repr__(self):
        attr_info = []
        if self._model_id:
            attr_info.append(f'model_id={self._model_id}')
        if self._model_version_id:
            attr_info.append(f'model_version_id={self._model_version_id}')
        if self._model_name:
            attr_info.append(f'name={self._model_name}')
        info_str = '\n  '.join(attr_info)
        return f'BasetenDeployedModel<\n  {info_str}\n>'

    @raises_api_error
    def log_metrics(self, metrics: dict) -> None:
        if not metrics:
            raise ValueError('metrics must be provided.')
        if not self._model_version_id and not self._model_id:
            raise ValueError('Please use on a BasetenDeployedModel instantiated with a model_version_id or model_id.')

        if self._model_id:
            api.create_model_metrics(metric_data=metrics, model_id=self._model_id)
        elif self._model_version_id:
            api.create_model_metrics(metric_data=metrics, model_version_id=self._model_version_id)
        logger.info('✅ Metrics logged successfully ✅')

    @raises_api_error
    def get_metrics(self) -> dict:
        if self._model_id:
            return api.model_metrics(self._model_id)
        elif self._model_version_id:
            return api.model_version_metrics(self._model_version_id)

        raise ValueError('Please use on a BasetenDeployedModel instantiated with a model_version_id or model_id.')

    @raises_api_error
    def log_config(self, config: dict) -> None:
        if not config:
            raise ValueError('config must be provided.')
        if not self._model_version_id and not self._model_id:
            raise ValueError('Please use on a BasetenDeployedModel instantiated with a model_version_id or model_id.')

        if self._model_id:
            api.create_model_configuration(configuration=config, model_id=self._model_id)
        elif self._model_version_id:
            api.create_model_configuration(configuration=config, model_version_id=self._model_version_id)
        logger.info('✅ Config logged successfully ✅')

    @raises_api_error
    def get_config(self) -> dict:
        if self._model_id:
            return api.model_configuration(self._model_id)
        elif self._model_version_id:
            return api.model_version_configuration(self._model_version_id)

        raise ValueError('Please use on a BasetenDeployedModel instantiated with a model_version_id or model_id.')

    @raises_api_error
    def add_artifact(
        self,
        name: str,
        list_of_files: List[str],
        description: str = None,
    ):
        """Create an artifact in the BaseTen backend to track

            Args:
                name (str): The name of the artifact to be created.
                list_of_files (List[str]): A list of files to be uploaded.
                description (str, optional): An optional description of the artifact
                model_id (str, optional): A model_id to associate with the artifact
                model_version_id (str, optional): A model_version_id to associate with the artifact
            Returns:
                BasetenArtifact

            Raises:
                ApiError: If there was an error communicating with the server.
            """
        from baseten.baseten_artifact import create_artifact
        if self._model_version_id:
            artifact = create_artifact(
                name, list_of_files, description=description, model_version_id=self._model_version_id
            )
        else:
            artifact = create_artifact(
                name, list_of_files, description=description, model_id=self._model_id
            )
        return artifact


@raises_api_error
def models_summary():
    return api.models_summary()



import configparser
import functools
import json
import logging
import os
import tarfile
import tempfile
from typing import IO, Dict, List

import requests
from colorama import Fore, Style
from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor
from tqdm import tqdm

from baseten.common import settings
from baseten.common.core import ARTIFACT_FILENAME, ApiError, AuthorizationError
from baseten.common.util import base64_encoded_json_str

logger = logging.getLogger(__name__)


def with_api_key(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        config = settings.read_config()
        try:
            api_key = config.get('api', 'api_key')
        except configparser.NoOptionError:
            raise AuthorizationError('You must first run the `baseten login` cli command.')
        result = func(api_key, *args, **kwargs)
        return result
    return wrapper


@with_api_key
def models(api_key):
    query_string = '''
    {
      models {
        id,
        name
      }
    }
    '''

    resp = _post_graphql_query(api_key, query_string)
    return resp['data']


@with_api_key
def models_summary(api_key):
    query_string = '''
    {
      models {
        id,
        name,
        description,
        versions {
          id
          created,
          configuration
        }
      }
    }
    '''

    resp = _post_graphql_query(api_key, query_string)
    return resp['data']['models']


@with_api_key
def create_model_from_scaffold(
    api_key,
    model_name,
    s3_key,
    model_framework,
    model_type,
    model_framework_req,
    dockerfile,
    build_args,
    semver_bump
):
    encoded_model_framework_req = base64_encoded_json_str(model_framework_req)
    encoded_build_args = base64_encoded_json_str(build_args)
    query_string = f'''
    mutation {{
      create_model_from_scaffold(name: "{model_name}"
                   s3_key: "{s3_key}",
                   model_framework: "{model_framework}",
                   model_type: "{model_type}",
                   semver_bump: "{semver_bump}",
                   encoded_model_framework_req: "{encoded_model_framework_req}",
                   dockerfile: "{dockerfile}"
                   encoded_build_args: "{encoded_build_args}"
      ) {{
        id,
        name,
        version_id
      }}
    }}
    '''

    resp = _post_graphql_query(api_key, query_string)
    return resp['data']['create_model_from_scaffold']


@with_api_key
def create_model_version_from_scaffold(
    api_key,
    model_id,
    s3_key,
    model_framework,
    model_type,
    model_framework_req,
    dockerfile,
    build_args,
    semver_bump
):
    encoded_model_framework_req = base64_encoded_json_str(model_framework_req)
    encoded_build_args = base64_encoded_json_str(build_args)
    query_string = f'''
    mutation {{
      create_model_version_from_scaffold(
                   model_id: "{model_id}"
                   s3_key: "{s3_key}",
                   model_framework: "{model_framework}",
                   model_type: "{model_type}",
                   semver_bump: "{semver_bump}",
                   encoded_model_framework_req: "{encoded_model_framework_req}",
                   dockerfile: "{dockerfile}"
                   encoded_build_args: "{encoded_build_args}"
      ) {{
        id,
        name,
        version_id
      }}
    }}
    '''

    resp = _post_graphql_query(api_key, query_string)
    return resp['data']['create_model_version_from_scaffold']


@with_api_key
def create_model(api_key,
                 model_name,
                 s3_key,
                 model_framework,
                 model_type,
                 input_shape,
                 feature_names,
                 feature_summary,
                 class_labels,
                 model_framework_req,
                 semver_bump,
                 model_files_dict):
    encoded_input_shape = base64_encoded_json_str(input_shape)
    encoded_feature_names = base64_encoded_json_str(feature_names)
    encoded_feature_summary = base64_encoded_json_str(feature_summary)
    encoded_class_labels = base64_encoded_json_str(class_labels)
    encoded_model_framework_req = base64_encoded_json_str(model_framework_req)
    encoded_model_files_dict = base64_encoded_json_str(model_files_dict)
    if not model_name:
        import coolname
        model_name = coolname.generate_slug(2)
    query_string = f'''
    mutation {{
      create_model(name: "{model_name}"
                   s3_key: "{s3_key}",
                   model_framework: "{model_framework}",
                   model_type: "{model_type}",
                   semver_bump: "{semver_bump}",
                   encoded_input_shape: "{encoded_input_shape}",
                   encoded_feature_names: "{encoded_feature_names}",
                   encoded_feature_summary: "{encoded_feature_summary}",
                   encoded_class_labels: "{encoded_class_labels}",
                   encoded_model_framework_req: "{encoded_model_framework_req}"
                   encoded_model_files_dict: "{encoded_model_files_dict}"
      ) {{
        id,
        name,
        version_id
      }}
    }}
    '''

    resp = _post_graphql_query(api_key, query_string)
    return resp['data']['create_model']


@with_api_key
def models_versions(api_key, model_id):
    query_string = f'''
    {{
      model(id: "{model_id}") {{
        versions {{
          id,
          s3_key,
          feature_names,
          feature_summary
        }}
      }}
    }}
    '''

    resp = _post_graphql_query(api_key, query_string)
    return resp['data']['model']


@with_api_key
def model_version_input_shape(api_key, model_version_id):
    query_string = f'''
    {{
      model_version(id: "{model_version_id}") {{
        input_shape
      }}
    }}
    '''

    resp = _post_graphql_query(api_key, query_string)
    return resp['data']['model_version']['input_shape']


@with_api_key
def create_model_version(api_key,
                         model_id,
                         s3_key,
                         model_framework,
                         model_type,
                         input_shape,
                         feature_names,
                         feature_summary,
                         class_labels,
                         model_framework_req,
                         semver_bump,
                         model_files_dict):
    encoded_input_shape = base64_encoded_json_str(input_shape)
    encoded_feature_names = base64_encoded_json_str(feature_names)
    encoded_feature_summary = base64_encoded_json_str(feature_summary)
    encoded_class_labels = base64_encoded_json_str(class_labels)
    encoded_model_framework_req = base64_encoded_json_str(model_framework_req)
    encoded_model_files_dict = base64_encoded_json_str(model_files_dict)
    query_string = f'''
    mutation {{
      create_model_version(model_id: "{model_id}",
                           s3_key: "{s3_key}",
                           model_framework: "{model_framework}",
                           model_type: "{model_type}",
                           semver_bump: "{semver_bump}",
                           encoded_input_shape: "{encoded_input_shape}",
                           encoded_feature_names: "{encoded_feature_names}",
                           encoded_feature_summary: "{encoded_feature_summary}",
                           encoded_class_labels: "{encoded_class_labels}",
                           encoded_model_framework_req: "{encoded_model_framework_req}"
                           encoded_model_files_dict: "{encoded_model_files_dict}") {{
        id,
        s3_key,
        feature_names,
        feature_summary
      }}
    }}
    '''

    resp = _post_graphql_query(api_key, query_string)
    return resp['data']['create_model_version']


@with_api_key
def create_model_configuration(
    api_key: str,
    configuration: dict,
    model_id: str = None,
    model_version_id: str = None,
):
    # encode for None/null
    encoded_configuration = base64_encoded_json_str(configuration)

    model_identifier_field = None
    if model_id:
        model_identifier_field = f'model_id: "{model_id}"'
    elif model_version_id:
        model_identifier_field = f'model_version_id: "{model_version_id}"'
    else:
        raise ValueError('Either model_version_id or model_id must be specified.')

    query_string = f'''
    mutation {{
        create_model_configuration(
            encoded_configuration: "{encoded_configuration}",
            {model_identifier_field}
        )
        {{
            configuration_id: id,
        }}
    }}
    '''
    resp = _post_graphql_query(api_key, query_string)
    return resp['data']['create_model_configuration']


@with_api_key
def model_version_configuration(api_key, model_version_id):
    query_string = f'''
    {{
      model_version(id: "{model_version_id}") {{
        configuration
      }}
    }}
    '''

    resp = _post_graphql_query(api_key, query_string)
    return resp['data']['model_version']['configuration']


@with_api_key
def model_configuration(api_key, model_id):
    query_string = f'''
    {{
      model(id: "{model_id}") {{
        configuration
      }}
    }}
    '''

    resp = _post_graphql_query(api_key, query_string)
    return resp['data']['model']['configuration']


@with_api_key
def create_model_metrics(
    api_key: str,
    metric_data: dict,
    model_id: str = None,
    model_version_id: str = None,
):
    # encode for None/null
    encoded_metric_data = base64_encoded_json_str(metric_data)

    model_identifier_field = None
    if model_id:
        model_identifier_field = f'model_id: "{model_id}"'
    elif model_version_id:
        model_identifier_field = f'model_version_id: "{model_version_id}"'
    else:
        raise ValueError('Either model_version_id or model_id must be specified.')

    query_string = f'''
    mutation {{
        create_model_metrics(
            encoded_metric_data: "{encoded_metric_data}",
            {model_identifier_field}
        )
        {{
            metrics_id: id,
        }}
    }}
    '''
    resp = _post_graphql_query(api_key, query_string)
    return resp['data']['create_model_metrics']


@with_api_key
def model_version_metrics(api_key, model_version_id):
    query_string = f'''
    {{
      model_version(id: "{model_version_id}") {{
        metrics
      }}
    }}
    '''

    resp = _post_graphql_query(api_key, query_string)
    return resp['data']['model_version']['metrics']


@with_api_key
def model_metrics(api_key, model_id):
    query_string = f'''
    {{
      model(id: "{model_id}") {{
        metrics
      }}
    }}
    '''

    resp = _post_graphql_query(api_key, query_string)
    return resp['data']['model']['metrics']


@with_api_key
def signed_s3_upload_post(api_key, model_file_name):
    query_string = f'''
    {{
      signed_s3_upload_url(model_file_name: "{model_file_name}") {{
        url,
        form_fields {{
          key,
          aws_access_key_id,
          policy,
          signature,
        }}
      }}
    }}
    '''

    resp = _post_graphql_query(api_key, query_string)
    return resp['data']['signed_s3_upload_url']


@with_api_key
def artifact_signed_s3_upload_post(api_key, file_name, file_size):
    query_string = f'''
    {{
      artifact_signed_s3_upload_url(file_name: "{file_name}", file_size: {file_size}) {{
        url,
        form_fields {{
          key,
          aws_access_key_id,
          policy,
          signature,
        }}
      }}
    }}
    '''

    resp = _post_graphql_query(api_key, query_string)
    return resp['data']['artifact_signed_s3_upload_url']


@with_api_key
def register_data_for_model(api_key, s3_key, model_version_id, data_name):
    query_string = f'''
    mutation {{
        create_sample_data_file(model_version_id: "{model_version_id}",
                                name: "{data_name}",
                                s3_key: "{s3_key}") {{
          id
        }}
    }}
    '''
    resp = _post_graphql_query(api_key, query_string)
    return resp['data']['create_sample_data_file']


@with_api_key
def create_artifact(
    api_key: str,
    s3_key: str,
    name: str,
    description: str = None,
    model_id: str = None,
    model_version_id: str = None,
):
    # encode for None/null
    description = base64_encoded_json_str(description)
    model_id = base64_encoded_json_str(model_id)
    model_version_id = base64_encoded_json_str(model_version_id)

    query_string = f'''
    mutation {{
        create_artifact(
            name: "{name}",
            s3_key: "{s3_key}",
            encoded_description: "{description}",
            encoded_model_id: "{model_id}",
            encoded_model_version_id: "{model_version_id}"
        )
        {{
            artifact_id: id,
            name,
            description
        }}
    }}
    '''
    resp = _post_graphql_query(api_key, query_string)
    return resp['data']['create_artifact']


@with_api_key
def create_artifact_link(
    api_key: str,
    artifact_id: str,
    model_id: str = None,
    model_version_id: str = None,
):
    # encode for None/null
    model_id = base64_encoded_json_str(model_id)
    model_version_id = base64_encoded_json_str(model_version_id)

    query_string = f'''
    mutation {{
        link_artifact(
            id: "{artifact_id}",
            encoded_model_id: "{model_id}",
            encoded_model_version_id: "{model_version_id}"
        )
        {{
            id
        }}
    }}
    '''
    resp = _post_graphql_query(api_key, query_string)
    return resp['data']['link_artifact']


@with_api_key
def delete_artifact(
    api_key: str,
    artifact_id: str,
):
    query_string = f'''
    mutation {{
        delete_artifact(
            artifact_id: "{artifact_id}",
        )
        {{
            ok
        }}
    }}
    '''
    resp = _post_graphql_query(api_key, query_string)
    return resp['data']['delete_artifact']


@with_api_key
def update_artifact(
    api_key: str,
    artifact_id: str,
    name: str = None,
    description: str = None,
    s3_key: str = None,
):
    update_str = ''
    if name:
        update_str += f'name: "{name}",\n'
    if description:
        update_str += f'description: "{description}",\n'
    if s3_key:
        update_str += f's3_key: "{s3_key}",\n'

    query_string = f'''
    mutation {{
        update_artifact(
            artifact_id: "{artifact_id}",
            {update_str}
        )
        {{
            id
            name
            description
        }}
    }}
    '''
    resp = _post_graphql_query(api_key, query_string)
    return resp['data']['update_artifact']


@with_api_key
def artifact_url(
    api_key: str,
    artifact_id: str,
):
    query_string = f'''{{
        artifact(
            id: "{artifact_id}",
        )
        {{
            id, url
        }}
    }}
    '''
    resp = _post_graphql_query(api_key, query_string)
    return resp['data']['artifact']['url']


@with_api_key
def artifact_links(
    api_key: str,
    artifact_id: str,
):
    query_string = f'''{{
        artifact(
            id: "{artifact_id}",
        )
        {{
            model_ids: oracle_ids
            model_version_ids: oracle_version_ids
        }}
    }}
    '''
    resp = _post_graphql_query(api_key, query_string)
    return resp['data']['artifact']


@with_api_key
def model_linked_artifacts(
    api_key: str,
    model_id: str,
):
    query_string = f'''{{
        model(
            id: "{model_id}",
        )
        {{
            artifacts {{
                artifact_id: id,
                name,
                description
            }}
        }}
    }}
    '''
    resp = _post_graphql_query(api_key, query_string)
    return resp['data']['model']['artifacts']


@with_api_key
def model_version_linked_artifacts(
    api_key: str,
    model_version_id: str,
):
    query_string = f'''{{
        model_version(
            id: "{model_version_id}",
        )
        {{
            artifacts {{
                artifact_id: id,
                name,
                description
            }}
        }}
    }}
    '''
    resp = _post_graphql_query(api_key, query_string)
    return resp['data']['model_version']['artifacts']


@with_api_key
def predict_for_model(api_key, model_id: str, inputs: list, metadata: List[Dict] = None) -> List[List]:
    """Call the model's predict given the input json.

    Args:
        api_key (str)
        model_id (str)
        inputs (list)
        metadata (List[Dict]): Metadata key/value pairs (e.g. name, url), one for each input.

    Raises:
        RequestException: If there was an error communicating with the server.
    """
    predict_url = f'{settings.API_URL_BASE}/models/{model_id}/predict'
    return _predict(api_key, predict_url, inputs, metadata)


@with_api_key
def predict_for_model_version(api_key, model_version_id: str, inputs: list, metadata: List[Dict] = None) -> List[List]:
    """Call the model version's predict given the input json.

    Args:
        api_key (str)
        model_version_id (str)
        inputs (list)
        metadata (List[Dict]): Metadata key/value pairs (e.g. name, url), one for each input.

    Raises:
        RequestException: If there was an error communicating with the server.
    """
    predict_url = f'{settings.API_URL_BASE}/model_versions/{model_version_id}/predict'
    return _predict(api_key, predict_url, inputs, metadata)


@with_api_key
def set_primary(api_key, model_version_id: str):
    """Promote this version of the model as the primary version.

    Args:
        api_key (str)
        model_version_id (str)
    """
    query_string = f'''
    mutation {{
      update_model_version(model_version_id: "{model_version_id}", is_primary: true) {{
        id,
        is_primary,
      }}
    }}
    '''
    resp = _post_graphql_query(api_key, query_string)
    return resp['data']['update_model_version']


@with_api_key
def update_model_features(api_key, model_version_id: str, feature_names: list, class_labels: list = None):
    """Update the feature names for the model.

    Args:
        api_key (str)
        model_version_id (str)
        feature_names (list)
        class_labels (Optional[list]): applies only to classifiers.
    """
    encoded_feature_names = base64_encoded_json_str(feature_names)
    encoded_class_labels = base64_encoded_json_str(class_labels)
    query_string = f'''
    mutation {{
      update_model_version(model_version_id: "{model_version_id}",
                           encoded_feature_names: "{encoded_feature_names}",
                           encoded_class_labels: "{encoded_class_labels}") {{
        id,
        feature_names,
      }}
    }}
    '''
    resp = _post_graphql_query(api_key, query_string)
    return resp['data']['update_model_version']


@with_api_key
def install_requirements(api_key, requirements_txt):
    escaped_requirements_txt = requirements_txt.replace('\n', '\\n')  # Otherwise the mutation becomes invalid graphql.
    query_string = f'''
    mutation {{
      create_pynode_requirement(requirements_txt: "{escaped_requirements_txt}") {{
        id
        status
        error_message
      }}
    }}
    '''
    resp = _post_graphql_query(api_key, query_string)
    return resp['data']['create_pynode_requirement']


@with_api_key
def requirement_status(api_key, requirement_id):
    query_string = f'''
    {{
      pynode_requirement(id: "{requirement_id}") {{
        id
        status
        error_message
      }}
    }}
    '''

    resp = _post_graphql_query(api_key, query_string)
    return resp['data']['pynode_requirement']


def _predict(api_key, predict_url: str, inputs: list, metadata: List[Dict] = None) -> List[List]:
    resp = _post_rest_query(api_key, predict_url, {'inputs': inputs, 'metadata': metadata})
    resp_json = json.loads(resp.content)
    return resp_json['predictions']


def _headers(api_key):
    return {'Authorization': f'Api-Key {api_key}'}


def _post_graphql_query(api_key, query_string) -> dict:
    resp = requests.post(f'{settings.API_URL_BASE}/graphql/', data={'query': query_string}, headers=_headers(api_key))
    if not resp.ok:
        logger.error(f'GraphQL endpoint failed with error: {resp.content}')
        resp.raise_for_status()
    resp_dict = resp.json()
    errors = resp_dict.get('errors')
    if errors:
        raise ApiError(errors[0]['message'], resp)
    return resp_dict


def _post_rest_query(api_key, url, post_body_dict):
    resp = requests.post(url, json=post_body_dict, headers=_headers(api_key))
    resp.raise_for_status()
    return resp


def upload_model(serialize_file: IO, file_ext: str, file_name: str) -> str:
    """Uploads the serialized model to the appropriate environment

    Args:
        serialize_file (file): A file-like object that is the serialized representation of the model object.
        file_ext (str): The file extension for the saved model

    Returns:
        str: The key for the uploaded model

    Raises:
        RequestException: If there was an error communicating with the server.
    """

    model_file_name = f'{file_name}.{file_ext}'
    this_signed_s3_upload_post = signed_s3_upload_post(model_file_name)
    logger.debug(f'Signed s3 upload post:\n{json.dumps(this_signed_s3_upload_post, indent=4)}')

    form_fields = this_signed_s3_upload_post['form_fields']
    form_fields['AWSAccessKeyId'] = form_fields.pop('aws_access_key_id')  # S3 expects key name AWSAccessKeyId
    form_fields['file'] = (file_name, serialize_file)
    logger.info('🚀 Uploading model to BaseTen 🚀')

    return _upload_file(this_signed_s3_upload_post, form_fields)


def upload_artifact(serialize_file: IO, file_ext: str, file_name: str, file_size: int) -> str:
    """Uploads the serialized artifact to the appropriate environment

    Args:
        serialize_file (file): A file-like object that is the serialized representation of the artifact object.
        file_ext (str): The file extension for the saved artifact

    Returns:
        str: The key for the uploaded artifact

    Raises:
        RequestException: If there was an error communicating with the server.
    """
    artifact_file_name = f'{file_name}.{file_ext}'
    this_signed_s3_upload_post = artifact_signed_s3_upload_post(artifact_file_name, file_size)
    logger.debug(f'Signed s3 upload post:\n{json.dumps(this_signed_s3_upload_post, indent=4)}')

    form_fields = this_signed_s3_upload_post['form_fields']
    form_fields['AWSAccessKeyId'] = form_fields.pop('aws_access_key_id')  # S3 expects key name AWSAccessKeyId
    form_fields['file'] = (file_name, serialize_file)
    logger.info('🚀 Uploading artifact to BaseTen 🚀')

    return _upload_file(this_signed_s3_upload_post, form_fields)


def _upload_file(this_signed_s3_upload_post: dict, form_fields: dict) -> str:
    encoder = MultipartEncoder(fields=form_fields)
    encoder_len = encoder.len
    pbar = tqdm(total=encoder_len, unit_scale=True,
                bar_format="Upload Progress: {percentage:3.0f}%% |%s{bar:100}%s| {n_fmt}/{total_fmt}"
                % (Fore.BLUE, Fore.RESET))

    def callback(monitor):
        progress = monitor.bytes_read - pbar.n
        pbar.update(progress)

    monitor = MultipartEncoderMonitor(encoder, callback)
    resp = requests.post(
        this_signed_s3_upload_post['url'], data=monitor, headers={'Content-Type': monitor.content_type}
    )
    resp.raise_for_status()
    pbar.close()
    logger.info('🔮 Upload successful!🔮')

    logger.debug(f'File upload HTTP status code: {resp.status_code} and content:\n{resp.content}')

    return this_signed_s3_upload_post['form_fields']['key']


def _tar_a_list_of_files(list_of_files: List[str]) -> tempfile.NamedTemporaryFile:
    temp_file = tempfile.NamedTemporaryFile(suffix='.tgz')
    with tarfile.open(temp_file.name, 'w:gz') as tar:
        for file in list_of_files:
            tar.add(file)
    temp_file.file.seek(0)
    return temp_file


def serialize_artifact_to_s3(name, list_of_files):
    logger.info(f'Serializing {Fore.BLUE}{name}{Style.RESET_ALL} artifact.')
    artifact_tgz = _tar_a_list_of_files(list_of_files)
    logger.info('Making contact with BaseTen 👋 👽')
    s3_key = upload_artifact(artifact_tgz, 'tgz', ARTIFACT_FILENAME, os.path.getsize(artifact_tgz.name))
    return s3_key



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['baseten', 'baseten.common']

package_data = \
{'': ['*']}

install_requires = \
['baseten-scaffolding==0.0.11',
 'click>=7.0',
 'colorama>=0.4.3',
 'coolname>=1.1.0',
 'jinja2>=2.10.3',
 'joblib>=0.12.5',
 'pyyaml>=5.1',
 'requests-toolbelt>=0.9.1,<0.10.0',
 'requests>=2.22',
 'tqdm>=4.62.1,<5.0.0']

entry_points = \
{'console_scripts': ['baseten = baseten.cli:cli_group']}

setup_kwargs = {
    'name': 'baseten',
    'version': '0.1.20',
    'description': '',
    'long_description': None,
    'author': 'Amir Haghighat',
    'author_email': 'amir@baseten.co',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<3.11',
}


setup(**setup_kwargs)
